/*
 *	Title:		o3find
 *	Version:	0.8.1
 *	Created by:	Paolo Copello
 *	Home page:	http://web.tiscali.it/fanelia
 *	
 *	Copyright (C) 2003-2005 Paolo Copello
 *	
 *	******************************************
 *	This file is part of o3find
 *	******************************************
 *	
 *	o3find is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	any later version.
 *	
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	
 */




// suppress "truncated debug info" warnings (VC++ specific)
#ifdef _MSC_VER
#pragma warning (disable: 4786)
#pragma warning (disable: 4788)
#endif // _MSC_VER




#include <fstream>
//#include <sstream>
#include <iostream>

using namespace std;

#include "o3findclass.h"
#include "fnutils.h"
#include "xmlmemfile.h"

namespace Fanel
{


//class OoFinder

/** OoFinder constructor
	bCaseSensitive		false to ignore case
	bShowErrors			false to hide errors and warnings
	bShowNotFound		true to show both found and not-found files
	bBriefMode			true to hide the matching lines
	search_in			content/metadata/both
*/
OoFinder::OoFinder(bool bCaseSensitive, bool bShowErrors,
					bool bShowNotFound, bool bBriefMode,
					SEARCH_IN search_in, bool bReadAlreadyOpenFile,
					bool bIgnoreStyle)
: _bRunning(false), _bCaseSensitive(bCaseSensitive), _bShowErrors(bShowErrors),
	_bShowNotFound(bShowNotFound), _bBriefMode(bBriefMode), _search_in(search_in),
	_bReadAlreadyOpenFile(bReadAlreadyOpenFile), _bIgnoreStyle(bIgnoreStyle)
{
	Clear();

	_adaptMap.insert(AdaptMapType::value_type('&', "&amp;"));
	_adaptMap.insert(AdaptMapType::value_type('<', "&lt;"));
	_adaptMap.insert(AdaptMapType::value_type('>', "&gt;"));
	_adaptMap.insert(AdaptMapType::value_type('\'', "&apos;"));
	_adaptMap.insert(AdaptMapType::value_type('\"', "&quot;"));

#ifdef WIN32
	// read Windows version: special chars like accented letters 
	// are supported only on Windows 98 or later
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx (&osvi);
	_bAccentedLettersSupported = VER_PLATFORM_WIN32_NT || 
	   (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) &&
	   ( (osvi.dwMajorVersion > 4) ||
	   ( (osvi.dwMajorVersion == 4) && (osvi.dwMinorVersion > 0) ) );
#endif

}

OoFinder::~OoFinder()
{
}






/** Find the search string in another string 
	nLength is the length of str 
	Note: since I don't know if str is null-terminated, I can't use strstr

	@todo Portability!!!
*/
bool OoFinder::FindSubstring(const char* str, int nLength, const string& sToBeFound)
{
	int nPatternLen = sToBeFound.length();

#ifdef WIN32
	// WINDOWS CODE
	for(int k = 0; k <= nLength-nPatternLen; k++)
	{
		if(_bCaseSensitive)
		{
			if(0 == strncmp(str+k, sToBeFound.c_str(), nPatternLen))
				return true;
		}
		else
		{
			if(0 == _strnicmp(str+k, sToBeFound.c_str(), nPatternLen))
				return true;
		}
	}
#else
	// NO-WINDOWS CODE
	// TODO...
#endif

	return false;
}

/** Find the (Unicode) search string in another string */
bool OoFinder::FindSubstringW(const char* str, int nLength, const wstring& sToBeFound)
{
	int nPatternLen = sToBeFound.length();

#ifdef WIN32
	// WINDOWS CODE
	for(int k = 0; k <= nLength-nPatternLen; k++)
	{
		if(_bCaseSensitive)
		{
			if(0 == wcsncmp( (wchar_t*)(str+k), sToBeFound.c_str(), nPatternLen))
				return true;
		}
		else
		{
			if(0 == _wcsnicmp( (wchar_t*)(str+k), sToBeFound.c_str(), nPatternLen))
				return true;
		}
	}
#else
	// NO-WINDOWS CODE
	// TODO...
#endif

	return false;
}


/** Find the search string in an in-memory XML file */
void OoFinder::FindInXML(const string& sFile, OoXmlMemFile& memFile)
{
	memFile.SeekToBegin();
	bool bFound = false;
	bool bEnd = memFile.GetLength() <= 0;

	string sChunk; //, sUpperChunk;
	const char* pCurrentChunk = NULL;
	int nLenght = 0;
	//bool bFoundNow = false;

	if (!_bIgnoreStyle)
	{
		while( NULL!= (pCurrentChunk = memFile.ReadNextXmlChunk(nLenght)) )
		{
			if(FindSubstring(pCurrentChunk, nLenght, _sToFindInXml))
			{
				sChunk.erase(sChunk.begin(), sChunk.end());
				sChunk.assign(pCurrentChunk, pCurrentChunk+nLenght);
				ReportFound(sFile, sChunk);	
				bFound = true;

				// in brief mode, one match is enough
				if (_bBriefMode)
					break;
			}
		}
	}
	else
	{
		ZIP_ULONGLONG nMaximumFileSize = memFile.GetLength();
		if (nMaximumFileSize > 0)
		{
			char* pStrippedBuffer = new char[nMaximumFileSize];
			memset(pStrippedBuffer, 0, nMaximumFileSize);
			// Strip off XML tags!
			int nCurrentPosition = 0;
			while( NULL!= (pCurrentChunk = memFile.ReadNextXmlChunk(nLenght)) )
			{
				memcpy(pStrippedBuffer+nCurrentPosition, pCurrentChunk, nLenght);
				nCurrentPosition += nLenght;
			}
			// At the end of the loop, nCurrentPosition contains 
			// the actual size of the stripped buffer

			if(FindSubstring(pStrippedBuffer, nCurrentPosition, _sToFindInXml))
			{
				//sChunk.erase(sChunk.begin(), sChunk.end());
				//sChunk.assign(pCurrentChunk, pCurrentChunk+nLenght);
				sChunk = _sToFindInXml;

				ReportFound(sFile, sChunk);	
				bFound = true;
			}

			delete [] pStrippedBuffer;
		}

//		const int BUF_SEMI_SIZE = 512;
//		const int BUF_SIZE = BUF_SEMI_SIZE*2;
//
//		char buffer[BUF_SIZE];
//		memset(buffer, 0, BUF_SIZE);
//
//		input.read(buffer, BUF_SIZE);
//		if (FindSubstring(buffer, BUF_SIZE, _sToFindUnadapted)
//			|| FindSubstringW(buffer, BUF_SIZE, _sToFindNormalSearchUnicode) )
//		{
//			ReportFound(sFile, string());
//			input.close();
//			return;
//		}
//
//		while(!input.eof())
//		{
//			memcpy(buffer, buffer+BUF_SEMI_SIZE, BUF_SEMI_SIZE);
//			memset(buffer+BUF_SEMI_SIZE, 0, BUF_SEMI_SIZE);
//			input.read(buffer+BUF_SEMI_SIZE, BUF_SEMI_SIZE);
//
//			if (FindSubstring(buffer, BUF_SIZE, _sToFindUnadapted)
//				|| FindSubstringW(buffer, BUF_SIZE, _sToFindNormalSearchUnicode) )
//			{
//				ReportFound(sFile, string());
//				input.close();
//				return;
//			}
//		}

	}

	// loop finished. Found?
	if (!bFound && _bShowNotFound)
	{
		ReportNotFound(sFile);
	}
}



/** Internal function that finds the search string in a single SO/OOo document. */ 
void OoFinder::InternalFind(const string& sFile)
{
	try
	{
		CZipArchive zip;
		zip.Open(sFile.c_str(), CZipArchive::zipOpenReadOnly, 0, _bReadAlreadyOpenFile);

		OoXmlMemFile contentMemFile;
		OoXmlMemFile metaMemFile;

		int iMetaData = -1, iContent = -1;

		// FIRST find and extract XML files, THEN close the ZIP file...

		// CONTENT
		if(_search_in != META_ONLY)
		{
			iContent = zip.FindFile("content.xml", CZipArchive::ffNoCaseSens, true);
			if(iContent<0)
			{
				if (_bShowErrors)
					ReportError(sFile, "content not found");
			}
			else
				zip.ExtractFile(iContent, contentMemFile);
		}
		// METADATA
		if(_search_in != CONTENT_ONLY)
		{
			iMetaData = zip.FindFile("meta.xml", CZipArchive::ffNoCaseSens, true);
			if (iMetaData<0)
			{
				if(_bShowErrors)
					ReportError(sFile, "metadata not found");
			}
			else
				zip.ExtractFile(iMetaData, metaMemFile);
		}

		zip.Close();

		// ... and THEN perform search
		if(_search_in != META_ONLY && iContent >= 0)
		{
			FindInXML(sFile, contentMemFile);
		}
		if(_search_in != CONTENT_ONLY && iMetaData >= 0)
		{
			FindInXML(sFile, metaMemFile);
		}
	}
	catch(CZipException& e)
	{
		if (e.m_iCause == CZipException::badZipFile || e.m_iCause == CZipException::cdirNotFound)
		{
			NormalFindInFile(sFile);
		}
		else if (_bShowErrors) 
		{
			if (e.m_iCause == CZipException::badZipFile || e.m_iCause == CZipException::cdirNotFound)
			{
				ReportError(sFile, "invalid file format");
			}
			else
			{
				ReportError(sFile, e.GetErrorDescription());
			}
		}
	}
	catch(CZipBaseException& e)
	{
		if (_bShowErrors) 
			ReportError(sFile, e.what());
	}
	catch(...)
	{
		if (_bShowErrors) 
			ReportError(sFile, "unknown error occurred");
	}
}

/** Find the string sToFind in a single file. */
void OoFinder::Find(const string& sFile, const string& sToFind)
{
	list<string> lst;
	lst.push_back(sFile);
	Find(lst, sToFind);
}

/** Find the string sToFind in many files. */
void OoFinder::Find(const list<string> & fileList, const string& sToFind)
{
	_bRunning = true;
	Clear();

	_sToFindUnadapted = sToFind;

	AdaptStringToXml(sToFind, _sToFindInXml);	
	AdaptStringToUnicodeSearch(sToFind, _sToFindNormalSearchUnicode);

	/*if(!_bCaseSensitive)
	{
		Str::ToUpper(sToFind, _sToFindInXml);
	}
	else
	{
		_sToFindInXml = sToFind;
	}*/

	for(list<string>::const_iterator it = fileList.begin(); it != fileList.end(); ++it)
	{
		InternalFind(*it);

		if(!_bRunning)
			break;
	}
	_bRunning = false;
}

/*void OoFinder::GetResults(int& nScanned, int& nFound, int& nErrors, list<string>& results)
{

}*/

//const list<string>& OoFinder::GetResultList() const
//{
//	return _results;
//}


/** Not used at the moment */ 
void OoFinder::Clear()
{
//	_iScanned = 0;
//	_iErrors = 0;
//	_iFound = 0;
//	_results.clear();
}

/** Stop search. Can be used by multithreaded applications. */
void OoFinder::Stop()
{
	_bRunning = false;
}


/*virtual*/void OoFinder::AdaptStringToXml(const string& strIn, string& strOut)
{
	strOut.erase(strOut.begin(), strOut.end());
	strOut.reserve(strIn.size());
	// browse the input string and build the second using the map
	for(string::const_iterator it=strIn.begin(); it!= strIn.end(); ++it)
	{
		AdaptMapType::const_iterator found = _adaptMap.find(*it);
		if(found != _adaptMap.end())
		{
			strOut += found->second;
		}
		else
		{
			strOut += *it;
		}
	}

#ifdef WIN32
	if(_bAccentedLettersSupported)
	{
		size_t len = strOut.length()*2 +1;
		WCHAR* pwszTmpSearch = new WCHAR[len];
		char* pszConvertedSearch = new char[len] ;
		memset((void *)pwszTmpSearch, 0, len);
        memset((void *)pszConvertedSearch, 0, len);
		MultiByteToWideChar(CP_ACP, 0, strOut.c_str(), strOut.length()+1, pwszTmpSearch, len) ;
		WideCharToMultiByte(CP_UTF8, 0, pwszTmpSearch, -1, pszConvertedSearch, len, NULL, NULL );
		strOut = pszConvertedSearch;
		delete [] pwszTmpSearch;
		delete [] pszConvertedSearch;
	}
#else
	// NO-WINDOWS CODE
	// TODO...
#endif
}

/*virtual*/ void OoFinder::AdaptStringToUnicodeSearch(const string& strIn, wstring& strOut)
{
	string strTmp = strIn;
#ifdef WIN32
	if(_bAccentedLettersSupported)
	{
		size_t len = strTmp.length()*2 +1;
		WCHAR* pwszTmpSearch = new WCHAR[len];
		//char* pszConvertedSearch = new char[len] ;
		memset((void *)pwszTmpSearch, 0, len);
        //memset((void *)pszConvertedSearch, 0, len);
		MultiByteToWideChar(CP_ACP, 0, strTmp.c_str(), strTmp.length()+1, pwszTmpSearch, len) ;
		//WideCharToMultiByte(CP_UTF8, 0, pwszTmpSearch, -1, pszConvertedSearch, len, NULL, NULL );
		strOut = pwszTmpSearch;
		delete [] pwszTmpSearch;
		//delete [] pszConvertedSearch;
	}
#else
	// NO-WINDOWS CODE
	// TODO...
#endif
}


/*virtual*/void OoFinder::UnAdaptStringFromXml(const string& strIn, string& strOut)
{
	strOut.erase(strOut.begin(), strOut.end());
	strOut.reserve(strIn.length());
	// browse the string and build the output using the map
	int nPos = 0;
	for(string::const_iterator it=strIn.begin(); 
		it!= strIn.end(); ++it, ++nPos)
	{
		AdaptMapType::const_iterator mapIt = _adaptMap.begin();
		for(; mapIt != _adaptMap.end(); ++mapIt)
		{
			if(0 == strIn.compare(nPos, mapIt->second.length(), mapIt->second))
			{
				strOut += mapIt->first;
				it += mapIt->second.length()-1;
				nPos += mapIt->second.length()-1;
				break;
			}
		}
		if(mapIt == _adaptMap.end())
		{
			// not found in map
			strOut += *it;
		}
	}

#ifdef WIN32
	if(_bAccentedLettersSupported)
	{
		size_t len = strOut.length()*2 +1;
		WCHAR* pwszTmpSearch = new WCHAR[len];
		char* pszConvertedSearch = new char[len] ;
		memset((void *)pwszTmpSearch, 0, len);
        memset((void *)pszConvertedSearch, 0, len);
		MultiByteToWideChar(CP_UTF8, 0, strOut.c_str(), strOut.length()+1, pwszTmpSearch, len) ;
		WideCharToMultiByte(CP_OEMCP, WC_COMPOSITECHECK, pwszTmpSearch, -1, pszConvertedSearch, len, NULL, NULL );
		strOut = pszConvertedSearch;
		delete [] pwszTmpSearch;
		delete [] pszConvertedSearch;
	}
#else
	// NO-WINDOWS CODE
	// TODO...
#endif

}

//void OoFinder::ReportFound(const string& sFile, const string& sLine)
//{
//}

//void OoFinder::ReportError(const string& sFile, const string& sMessage)
//{
//}
//
//void OoFinder::ReportNotFound(const string& sFile)
//{
//}


void OoFinder::NormalFindInFile(const string& sFile)
{
	const int BUF_SEMI_SIZE = 512;
	const int BUF_SIZE = BUF_SEMI_SIZE*2;

	ifstream input(sFile.c_str(), ios_base::binary );

	char buffer[BUF_SIZE];
	memset(buffer, 0, BUF_SIZE);

	input.read(buffer, BUF_SIZE);
	if (FindSubstring(buffer, BUF_SIZE, _sToFindUnadapted)
		|| FindSubstringW(buffer, BUF_SIZE, _sToFindNormalSearchUnicode) )
	{
		ReportFound(sFile, string());
		input.close();
		return;
	}

	while(!input.eof())
	{
		memcpy(buffer, buffer+BUF_SEMI_SIZE, BUF_SEMI_SIZE);
		memset(buffer+BUF_SEMI_SIZE, 0, BUF_SEMI_SIZE);
		input.read(buffer+BUF_SEMI_SIZE, BUF_SEMI_SIZE);

		if (FindSubstring(buffer, BUF_SIZE, _sToFindUnadapted)
			|| FindSubstringW(buffer, BUF_SIZE, _sToFindNormalSearchUnicode) )
		{
			ReportFound(sFile, string());
			input.close();
			return;
		}
	}

	input.close();

	// loop finished. NOT Found
	if (_bShowNotFound)
	{
		ReportNotFound(sFile);
	}

}



}  // end namespace Fanel

